/*
 * initialization.c
 *
 * Created: 21/03/2014 10:15:46
 *  Author: marcadella
 */ 

#include "initialization.h"

// Constants definitions
#define FCPU_HZ 12000000
#define FPBA_HZ 12000000


/************************** Initialisation functions **************************/


// *********** Clock

static pcl_freq_param_t pcl_freq_param_custom = {
	.cpu_f = FCPU_HZ,
	.pba_f = FPBA_HZ,
	.osc0_f = FOSC0,
	.osc0_startup = OSC0_STARTUP
};

int pm_configure_clocks_optimized(pcl_freq_param_t *param)
	{
	#define PM_MAX_MUL                         ((1 << AVR32_SCIF_PLLMUL_SIZE) - 1)
	#define AVR32_PM_PBA_MAX_FREQ              66000000
	#define AVR32_PM_PLL_VCO_RANGE0_MAX_FREQ   240000000
	#define AVR32_PM_PLL_VCO_RANGE0_MIN_FREQ   160000000

	// Implementation for  UC3C parts.
	// Supported frequencies:
	// Fosc0 mul div PLL div2_en cpu_f pba_f   Comment
	//  12   15   1  192     1     12    12
	//  12    9   3   40     1     20    20    PLL out of spec
	//  12   15   1  192     1     24    12
	//  12    9   1  120     1     30    15
	//  12    9   3   40     0     40    20    PLL out of spec
	//  12   15   1  192     1     48    12
	//  12   15   1  192     1     48    24
	//  12    8   1  108     1     54    27
	//  12    9   1  120     1     60    15
	//  12    9   1  120     1     60    30
	//  12   10   1  132     1     66    16.5
	//
	unsigned long in_cpu_f  = param->cpu_f;
	unsigned long in_osc0_f = param->osc0_f;
	unsigned long mul, div, div2_en = 0, div2_cpu = 0, div2_pba = 0;
	unsigned long pll_freq, rest;
	bool b_div2_pba, b_div2_cpu;

	// Configure OSC0 in crystal mode, external crystal with a FOSC0 Hz frequency.
	scif_configure_osc_crystalmode(SCIF_OSC0, in_osc0_f);
	// Enable the OSC0
	scif_enable_osc(SCIF_OSC0, param->osc0_startup, true);
	// Set the main clock source as being OSC0.
	pm_set_mclk_source(PM_CLK_SRC_OSC0);

	// Start with CPU freq config
	if (in_cpu_f == in_osc0_f)
	{
		param->cpu_f = in_osc0_f;
		param->pba_f = in_osc0_f;
		return PASS;
	}
	else if (in_cpu_f < in_osc0_f)
	{
		// TBD
	}

	rest = in_cpu_f % in_osc0_f;

	for (div = 1; div < 32; div++)
	{
		if ((div * rest) % in_osc0_f == 0)
		break;
	}
	if (div == 32)
	return FAIL;

	mul = (in_cpu_f * div) / in_osc0_f;

	if (mul > PM_MAX_MUL)
	return FAIL;

	// export 2power from PLL div to div2_cpu
	while (!(div % 2))
	{
		div /= 2;
		div2_cpu++;
	}

	// Here we know the mul and div parameter of the PLL config.
	// . Check out if the PLL has a valid in_cpu_f.
	// . Try to have for the PLL frequency (VCO output) the highest possible value
	//   to reduce jitter.
	while (in_osc0_f * 2 * mul / div < AVR32_PM_PLL_VCO_RANGE0_MAX_FREQ)
	{
		if (2 * mul > PM_MAX_MUL)
		break;
		mul *= 2;
		div2_cpu++;
	}

	if (div2_cpu != 0)
	{
		div2_cpu--;
		div2_en = 1;
	}

	pll_freq = in_osc0_f * mul / (div * (1 << div2_en));

	// Update real CPU Frequency
	param->cpu_f = pll_freq / (1 << div2_cpu);
	mul--;

	scif_pll_opt_t opt;

	opt.osc = SCIF_OSC0,     // Sel Osc0 or Osc1
	opt.lockcount = 16,      // lockcount in main clock for the PLL wait lock
	opt.div = div,             // DIV=1 in the formula
	opt.mul = mul,             // MUL=7 in the formula
	opt.pll_div2 = div2_en,        // pll_div2 Divide the PLL output frequency by 2 (this settings does not change the FVCO value)
	opt.pll_wbwdisable = 0,  //pll_wbwdisable 1 Disable the Wide-Bandith Mode (Wide-Bandwith mode allow a faster startup time and out-of-lock time). 0 to enable the Wide-Bandith Mode.
	opt.pll_freq = (pll_freq < AVR32_PM_PLL_VCO_RANGE0_MIN_FREQ) ? 1 : 0,        // Set to 1 for VCO frequency range 80-180MHz, set to 0 for VCO frequency range 160-240Mhz.

	scif_pll_setup(SCIF_PLL0, &opt); // lockcount in main clock for the PLL wait lock

	/* Enable PLL0 */
	scif_pll_enable(SCIF_PLL0);

	/* Wait for PLL0 locked */
	scif_wait_for_pll_locked(SCIF_PLL0) ;

	rest = pll_freq;
	while (rest > AVR32_PM_PBA_MAX_FREQ ||
	rest != param->pba_f)
	{
		div2_pba++;
		rest = pll_freq / (1 << div2_pba);
		if (rest < param->pba_f)
		break;
	}

	// Update real PBA Frequency
	param->pba_f = pll_freq / (1 << div2_pba);


	if (div2_cpu)
	{
		b_div2_cpu = true;
		div2_cpu--;
	}
	else
	b_div2_cpu = false;

	if (div2_pba)
	{
		b_div2_pba = true;
		div2_pba--;
	}
	else
	b_div2_pba = false;

	if (b_div2_cpu == true )
	{
		pm_set_clk_domain_div(PM_CLK_DOMAIN_0, (pm_divratio_t) div2_cpu); // CPU
		pm_set_clk_domain_div(PM_CLK_DOMAIN_1, (pm_divratio_t) div2_cpu); // HSB
		pm_set_clk_domain_div(PM_CLK_DOMAIN_3, (pm_divratio_t) div2_cpu); // PBB
	}
	if (b_div2_pba == true )
	{
		pm_set_clk_domain_div(PM_CLK_DOMAIN_2, (pm_divratio_t) div2_pba); // PBA
		pm_set_clk_domain_div(PM_CLK_DOMAIN_4, (pm_divratio_t) div2_pba); // PBC
	}

	// Set Flashc Wait State
	flashc_set_flash_waitstate_and_readmode(param->cpu_f);

	// Set the main clock source as being PLL0.
	pm_set_mclk_source(PM_CLK_SRC_PLL0);

	return PASS;
}

void init_sys_clocks (void)
{
	if (pm_configure_clocks_optimized(&pcl_freq_param_custom) != 0) {
		while (1){
			cpu_relax();
		}
	}
}

